(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)

(* ebnf *)
(** abstract syntax of ebnf formulas *)

open Utils

type print_mode =
|  Normal
|  Short
|  Expand
|  Ignore

type ebnf = 
|	Eps    (* empty string *)
|	Void   (* empty set *)
|	Symb of string
|	Seq of ebnf list
|	Choice of ebnf list
|	Star of ebnf
|	Opt of ebnf

(** external view *)
type exp = ebnf

type t = (Grammar.symbol, exp * print_mode) Hashtbl.t

(* dump ebnf grammar *)
let rec string_of_ebnf eg ?(nopar = false) e = (
	let (op',cp') = if nopar then ("","") else ("( "," )") in
	match e with
	|	Void      -> "void"
	|	Eps       -> "eps"
	|	Symb    s -> (
		try (
			let (es, mode) = Hashtbl.find eg s in
			match mode with
			|	Expand -> string_of_ebnf eg ~nopar:nopar es
			|	_ -> s
		) with Not_found -> s
	)
	|	Seq    el -> string_of_ebnf_list eg "" el " " ""
	|	Choice el -> string_of_ebnf_list eg op' el " | " cp' 
	|	Star    e -> "{ "^(string_of_ebnf eg ~nopar:true e)^" }"
	|	Opt     e -> "[ "^(string_of_ebnf eg  ~nopar:true e)^" ]"
)
and string_of_ebnf_list eg op el sep cp  = (
	let sl = List.map (string_of_ebnf eg) el in
	string_of_string_list ~obr:op ~sep:sep ~cbr:cp sl 
)

(** normalization :
   - e <=> Void iff e = Void
   - e <=> Eps  iff e = Eps
   - e <=> Seq(e1, e2 ...) implies none of e1 are Seq

	args are supposed to be normalized
*)

let rec contains_eps = function
|  Void
		(* stupid: emptyset does not contain anything ! *)
      (* -> raise (Invalid_argument "\"Ebnf.contains_eps\" not defined for \"Void\"") *)
|  Symb    _ -> false
|  Seq    el -> List.fold_left (fun a e -> a && contains_eps e) true el
|  Choice el -> List.fold_left (fun a e -> a || contains_eps e) false el
|  Eps
|  Star    _
|  Opt     _ -> true



(* returns opt if needed *)
let make_opt e = (
	if contains_eps e then e
	else if e = Void then Eps
	else Opt e
)

(* Void neutral, if at least one Eps -> Opt *)
let make_choice (el: ebnf list) = (
	(* ce <=> contains eps *)
	let targ (a,ceps) e = (
		match e with
		|	Void -> (a,ceps)
		|	Eps -> (a,true)
		|  Opt (Choice el) -> (a @ el, true)
		|  Opt (_el) -> (a @ [e], true)
		|	Choice _el -> (a @ [e], ceps)
		|	_ -> (a @ [e], ceps)
	) in 
	match
		List.fold_left targ ([],false) el
	with
	|  ([],false) -> Void
	|  ([],true) -> Eps
	|  ([a],true) -> Opt a 
	|  ([a],false) -> a
	|  (args,true) -> Opt (Choice args)
	|  (args,false) -> Choice args 
)

exception Is_void

(* Void absorb, Eps neutral *)
let make_seq (el: ebnf list)  = (
	let targ a e = (
		match e with
		|	Void -> raise Is_void
		|	Eps -> a
		|	Seq el' -> a @ el'
		|	_ -> a @ [e]
	) in 
	try (
		let args = List.fold_left targ [] el in
		match args with
		| [] -> Eps
		| [a] -> a
		| _ -> Seq args
	) with Is_void -> Void
)

let dump eg ll groups = (
	let dump_def x = (
		let (e, mode) = Hashtbl.find eg x in
		match mode with
		|	Expand
		|	Ignore -> ()
		|	Normal
		|	Short -> (
			let pfx0 = Printf.sprintf "%s ::= " x in
			let pfx1 = Printf.sprintf "%s  |  " (String.make (String.length x) ' ') in
			let rl = match e with
			|	Choice l -> l
			|	_ -> [ e ]
			in
			(* top-level: one line per choice *)
			let pchoice pfx e = (
				Printf.printf "%s%s\n" pfx (string_of_ebnf eg ~nopar:true e);
				pfx1 
			) in
			let _ = List.fold_left pchoice pfx0 rl in
			Printf.printf "\n"
		)
	) in
	let dump_group (g,ll) = (
		Printf.printf "(* GROUP %s *)\n\n" g;
		List.iter dump_def ll
	) in
	if groups = [] then
		List.iter dump_def ll
	else
		List.iter dump_group groups

)


(* dump ebnf grammar -> latex (very similar) *)

(* command for terminal symbol (tokens) *)
let term_comm = "\\bnftoken"
(* command for terminal symbol (tokens) *)
let nonterm_left_comm = "\\bnfleftident"
let nonterm_right_comm = "\\bnfrightident"
let metadef = "\\bnfdef"
let metaor_comm = "\\bnfor"
let metaopt = "\\bnfopt"
let metalist = "\\bnflist"
let metagroup = "\\bnfgroup"

(* commands for (very) special characters, depends on tt or not *)
let rbrace = "\\bnfrightbrace"
let lbrace = "\\bnfleftbrace"
let greater = "\\bnfgreater"
let less = "\\bnfless"
let underscore = "\\bnfunderscore"
let backslash = "\\bnfbackslash"


let latex_prolog = "
\\begingroup\\makeatletter%
\\ifx"^metadef^"\\undefined%
\\gdef"^metadef^"{%
  {{$::=$}}%
  }%
\\fi%
\\ifx"^metaor_comm^"\\undefined%
\\gdef"^metaor_comm^"{%
  {{$\\,|\\,$}}%
  }%
\\fi%
\\ifx"^metaopt^"\\undefined%
\\gdef"^metaopt^"#1{%
  {{$\\lbrack$ #1 $\\rbrack$}}%
  }%
\\fi%
\\ifx"^metalist^"\\undefined%
\\gdef"^metalist^"#1{%
  {{$\\lbrace$ #1 $\\rbrace$}}%
  }%
\\fi%
\\ifx"^metagroup^"\\undefined%
\\gdef"^metagroup^"#1{%
  {{$($ #1 $)$}}%
  }%
\\fi%
\\ifx"^term_comm^"\\undefined%
\\gdef"^term_comm^"#1{%
  {{\\bfseries\\ttfamily #1}}%
  }%
\\fi%
\\ifx"^nonterm_left_comm^"\\undefined%
\\gdef"^nonterm_left_comm^"#1{%
  {{\\itshape #1}}%
  }%
\\fi%
\\ifx"^nonterm_right_comm^"\\undefined%
\\gdef"^nonterm_right_comm^"#1{%
  {{\\itshape #1}}%
  }%
\\fi%
\\gdef\\bnftt{cmtt}%
\\gdef\\bnfttselect#1#2{%
\\ifx\\f@family\\bnftt #1\\else #2\\fi%
}%
\\endgroup%

\\newcommand{"^underscore^"}{\\bnfttselect{\\char`\\_}{\\_}}
\\newcommand{"^backslash^"}{\\bnfttselect{\\char`\\\\}{\\textbackslash}}
\\newcommand{"^lbrace^"}{\\bnfttselect{\\char`\\{}{\\lbrace}}
\\newcommand{"^rbrace^"}{\\bnfttselect{\\char`\\}}{\\rbrace}}
\\newcommand{"^greater^"}{\\bnfttselect{\\char`\\>}{\\textgreater}}
\\newcommand{"^less^"}{\\bnfttselect{\\char`\\<}{\\textless}}
"

(*
For non term, there are 2 differents latex translation:
("toto", true) -> when appearing in left pos
("toto", false) -> when appearing in right pos

Ugly : _TKTAB shouldn't be a global var but a field in type t !
*)
let _TKTAB : (string * bool , string) Hashtbl.t = Hashtbl.create 100 

(* must see "large" because of ~ -> \textasciitilde{} *)
let buff = Buffer.create 128

(*
There is no way to define _, { and } > and < in such a way that it works
WHATEVER be the current font family, so we use a latex command
that can be redefined if needed.
The default commands works for tt family.
*)
let  texify s = (
	Buffer.reset buff ;
	let puts s = ( Buffer.add_string buff s) in
	let put c = ( Buffer.add_char buff c) in
	let escape c = ( Buffer.add_char buff '\\' ; put c) in

	let filter c = match c with
		|	'$'  -> escape c
		|	'&'  -> escape c
		|	'%'  -> escape c
		|	'#'  -> escape c
		|	'{'  -> puts (lbrace^"{}")
		|	'}'  -> puts (rbrace^"{}")
		|	'<'  -> puts (less^"{}")
		|	'>'  -> puts (greater^"{}")
		|	'_'  -> puts (underscore^"{}")
		|	'\\' -> puts (backslash^"{}")
		|	'^'  -> puts "\\textasciicircum{}"
		|	'~'  -> puts "\\textasciitilde{}"
		|	'@'  -> puts "@"
		|  _    -> put c
	in
	String.iter filter s ;
	Buffer.contents buff
)


let terminal s = term_comm^"{"^(texify s)^"}"
let nonterminal ?(left=false) s = (
	let l = if left then
		nonterm_left_comm
	else
		nonterm_right_comm
	in
	l^"{"^(texify s)^"}"
)

let metaor = " "^metaor_comm^"{} "

(*
   Specify that "id" is a terminal symbol:
   - should be replaced by "conc" when printed
   - should be "printed" in terminal style
   N.B. even simple char (e.g. +, * etc)
   must be declared with add_terminal in order
   to be printed in the right style
*)
let add_terminal (spec: string * string) : unit =
   let (tk,v) = spec in
   let  v' = terminal v in
   Hashtbl.add _TKTAB (tk,false) v'

let add_nonterminal (spec: string * string) : unit =
   let (tk,v) = spec in
   let  vl = nonterminal v ~left:true in
   Hashtbl.add _TKTAB (tk,true) vl;
   let  vr = nonterminal v ~left:false in
   Hashtbl.add _TKTAB (tk,false) vr

(*
   ebnf to latex 
*)
let platex = Printf.fprintf

let symb_to_latex ?(left = false) s = (
	let res = try (
		Hashtbl.find _TKTAB (s,left)
	) with Not_found -> (
		(* obsolete:
         - all nonterm MUST be in _TKTAB
         - when caching left  part, why not
           caching right part too !
		let (s',lft) =
         before first call of symb_to_latex
      try (
			Scanf.sscanf s "%C"
				(fun c -> terminal (String.make 1 c)),
			false
		) with _ ->
      nonterminal ~left:left s, left
		in
		Hashtbl.add _TKTAB (s,lft) s';
		s'
      *)
      add_nonterminal (s,s);
      Hashtbl.find _TKTAB (s,left)
	) in
	res
) 

let rec latex_of_ebnf eg ?(nopar = false) (e:ebnf) = (
	match e with
	|	Symb    s -> (
		try (
			let (es, mode) = Hashtbl.find eg s in
			match mode with
			|	Expand -> latex_of_ebnf eg ~nopar:nopar es
			|	_ -> symb_to_latex s
		) with Not_found -> symb_to_latex s
	)
	|	Seq    el -> latex_of_ebnf_list eg el " "
	|	Choice el -> (
		if nopar then
			latex_of_ebnf_list eg el metaor 
		else
			metagroup^"{"^(latex_of_ebnf_list eg el metaor)^"}"
	)
	|	Star    e -> metalist^"{"^(latex_of_ebnf eg ~nopar:true e)^"}"
	|	Opt     e -> metaopt^"{"^(latex_of_ebnf eg ~nopar:true e)^"}"
	|	Void -> "VOID"
	|	Eps -> "EPS"
)
and latex_of_ebnf_list eg (el:ebnf list) (sep:string)  = (
	let sl = List.map (latex_of_ebnf eg) el in
	string_of_string_list ~obr:"" ~sep:sep ~cbr:"" sl 
)


let defs_to_latex oc eg ll = (
	let def_to_latex x = (
		let (e, mode) = Hashtbl.find eg x in
			match mode with
			|	Ignore
			|	Expand -> ()
			|	_ -> (
(* Printf.fprintf stderr "def_to_latex %s short=%b\n" x short; *)
			let pfx0 = Printf.sprintf "%s & %s &" (symb_to_latex ~left:true x) metadef in
			let pfx1 = Printf.sprintf "%s & %s &" (String.make (String.length x) ' ') metaor in
			let rl = match e with
			|	Choice l -> if (mode = Short) then [ e ] else l
			|	_ -> [ e ]
			in
			(* top-level, not short: one line per choice *)
			let pchoice pfx e = (
				platex oc "%s\n" pfx;
				platex oc "  \\begin{minipage}[t]{\\rulerhs}\n";
				platex oc "  \\raggedright\n";
				platex oc "  %s\n" (latex_of_ebnf eg ~nopar:true e);
				platex oc "  \\end{minipage}\\\\\n";
				pfx1 
			) in
			let _ = List.fold_left pchoice pfx0 rl in
			platex oc "\n"
		)
	) in
	List.iter def_to_latex ll
)

(* max = the longest string expected in left part *)
let start_table oc max = (
	(* prologue *)
   (* BUG ! dont forget to texify max ! *)
	platex oc "\\settowidth\\rulelhs{%s}\n" (texify max);
	platex oc "\\settowidth\\rulemid{::=}\n" ;
	platex oc "\\setlength\\rulerhs{\\textwidth}\n" ;
	platex oc "\\addtolength\\rulerhs{-\\rulelhs}\n" ;
	platex oc "\\addtolength\\rulerhs{-\\rulemid}\n" ;
	platex oc "\\addtolength\\rulerhs{-6\\tabcolsep}\n\n" ;
	platex oc "\\begin{longtable}{lrl}\n";
)

let end_table oc = (
	platex oc "\\end{longtable}\n";
)

let longest_left ll = (
	let chmax (lm,m) x = (
		let lx = String.length x in
		if lx > lm then (lx,x)
		else (lm,m) 
	) in
	let (_lmax,max) = List.fold_left chmax (0,"") ll in
	max
)


let group_to_latex oc gnme eg ll = (
	(* max left-part width *)
	let max = longest_left ll in
	platex oc "\\newcommand{\\%s}{\n" gnme;
	start_table oc max;
	defs_to_latex oc eg ll ;
	end_table oc;
	platex oc "}\n";
)

let create : unit -> (string, (ebnf * print_mode)) Hashtbl.t =
	fun _ -> Hashtbl.create 10

let to_latex 
   (oc: out_channel)
	(eg: (string, (ebnf * print_mode)) Hashtbl.t)
	(ll: string list)
	(groups : (string * string list) list)
= (
   
	(* general prologue *)
	platex oc "%s" latex_prolog;
	platex oc "\\newlength\\rulelhs\n" ;
	platex oc "\\newlength\\rulemid\n" ;
	platex oc "\\newlength\\rulerhs\n" ;

	let wholetab = "ebnftable" in

	if (groups = [] ) then (
		(* single group named ebnftable *)
		group_to_latex oc wholetab eg ll ;
		(* call it for backward compatibility *)
		(* platex oc "\\%s\n" g ; *)
	) else (
		(* generates one command per group *)
		let do_group (gnme, ll) = (
			group_to_latex oc gnme eg ll
		) in
		List.iter do_group groups;
		let call_group (gnme, _ll) = (
			platex oc "\\paragraph{Ebnf group {\\it %s}}~\\\\\n" gnme;
			platex oc "\\%s\n\n" gnme;
		) in
		platex oc "\\newcommand{\\%s}{\n" wholetab;
		List.iter call_group groups;
		platex oc "}\n";
	)
)

let make_latex_top (oc: out_channel) (toincl:string) : unit = (

output_string oc "
%
% minimal tex main file for yacc2tex ebnf generated grammars
%
%

\\documentclass[11pt,a4paper,twoside]{article}

\\usepackage{color}
\\usepackage{hyperref}
\\usepackage{longtable}
\\usepackage{tabularx}
\\usepackage[pdftex]{graphicx}

\\usepackage{bold-extra}

%\\input{x11colors}


\\setlength{\\textwidth}{161mm} %21cm - (2 * 1cm)
\\setlength{\\textheight}{224mm}
\\setlength{\\oddsidemargin}{0pt}
\\setlength{\\evensidemargin}{0pt}
%\\setlength{\\topmargin}{0pt}
%\\setlength{\\partopsep}{0pt}
\\setlength{\\parskip}{0pt}
\\setlength{\\parindent}{0pt}

\\begin{document}

%
\\newcommand{\\bnftoken}[1]{{\\bfseries\\ttfamily #1}}
\\newcommand{\\bnfleftident}[1]{{\\color{blue}\\itshape #1}}
\\newcommand{\\bnfdef}{\\mbox{\\color{red}$::=$}}
\\newcommand{\\bnfor}{\\mbox{\\color{red}$|$}}
\\newcommand{\\bnfopt}[1]{\\mbox{\\color{red}$\\lbrack$} #1 \\mbox{\\color{red}$\\rbrack$}}
\\newcommand{\\bnflist}[1]{\\mbox{\\color{red}$\\lbrace$} #1 \\mbox{\\color{red}$\\rbrace$}}
\\newcommand{\\bnfgroup}[1]{\\mbox{\\color{red}$($} #1 \\mbox{\\color{red}$)$}}
\\input{" ;  output_string oc toincl; output_string oc "}
\\ebnftable

\\end{document}
"
)
