(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)

(* formal language grammar data type
   + associated functions
   very simple:
   - no "type" difference between terminal/non terminal
     they are all simple strings
     (basically a non-terminal is just a defined symbol)
*)

open Utils

type symbol = string

(* single rule = sequence of symbols *)
type rule = symbol list

(* for extern listing only *)
type pragma = (string * string * string)

type pragmas = (string * string, string) Hashtbl.t

let add_pragma_to_pragmas (p:pragmas) (f,n,v) =
   Hashtbl.replace p (f,n) v

type definition = {
   mutable def_left: string;
   mutable def_rules: rule list;
   def_pragmas: pragmas;
}

let add_rule (d:definition) (r:rule) =
   d.def_rules <- r::d.def_rules

let add_pragma_to_definition (d:definition) p =
   add_pragma_to_pragmas d.def_pragmas p 

let get_definition_pragma (d:definition) (f,n) =
   Hashtbl.find d.def_pragmas (f,n)

let dummy_definition = {
   def_left = "";
   def_rules = [];
   def_pragmas = Hashtbl.create 0;
}

let create_definition i pgs = {
   def_left = i;
   def_rules = [];
   def_pragmas = Hashtbl.copy pgs;
}

(*
   
*)
type t = {
   gr_tab : (string, definition) Hashtbl.t;
   mutable gr_def_list : string list;
   (* use a set just for fun ... *)
   mutable gr_lit_set : StringSet.t;
}

let create_grammar () = {
   gr_tab = Hashtbl.create 200;
   gr_def_list = [];
   gr_lit_set = StringSet.empty;
}

let find_definition g lf =
   Hashtbl.find g.gr_tab lf

(* return existing eq. or create a fresh empty one
   adding pragmas
*)
let add_definition (g:t) (id:string) (pgs:pragmas) : definition =
   try (
      find_definition g id
   ) with Not_found -> (
      let eq = create_definition id pgs in
      Hashtbl.add g.gr_tab id eq;
      g.gr_def_list <- id::(g.gr_def_list);
      eq
   )

let get_symbol_pragma g i (f,n) =
   let d = find_definition g i in
   get_definition_pragma d (f,n)

let symbol_rules g s =
   let eq = find_definition g s in eq.def_rules

let symbol_pragmas g s =
   let eq = find_definition g s in
   let gprs (f,n) v acc = (f,n,v)::acc in
   Hashtbl.fold gprs eq.def_pragmas []


let def_symbols g = g.gr_def_list
let lit_symbols g = StringSet.elements (g.gr_lit_set)


(*-----------------------------------------------
Parse a Ylexer token stream
(LL1 grammar)
-------------------------------------------------

grammar  : TK_end
         | TK_pragma grammar
         | TK_ident TK_colon [ TK_pragma ] rulelist grammar

rulelist : rule TK_bar rulelist
         | rule TK_semicol

rule     : eps 
         | TK_ident rule
         | TK_literal rule
         | TK_directive rule

-----------------------------------------------*)

(* type token = Ylexer.token *)
open Ylexer

let of_yacc (lexer : Ylexer.t) = (
   (* ugly but confortable:
      grammar is built by side effects ...
   *)
   let zegram = create_grammar () in

   (* global pragmas: tab of (family, name) -> value
      reading a (f,n,v') after (f,n,v) erases the v !
   *)
   let global_pragmas = Hashtbl.create 10 in
   let add_global_pragmas (f,n,v) =
(* Printf.fprintf stderr "GLOBAL PRAGMA: %s:%s=%s\n" f n v; *)
      add_pragma_to_pragmas global_pragmas (f,n,v)
   in

   (* current equation *)
   let current_definition = ref dummy_definition in
   let current_rule = ref [] in
   let add_to_rule i =
      current_rule := i::!current_rule
   in
   let add_literal l =
      zegram.gr_lit_set <- StringSet.add l zegram.gr_lit_set
   in
   let new_equation i = 
      let def = add_definition zegram i global_pragmas in
      current_definition := def 
   in
   let add_local_pragma (f,n,v) =
(* Printf.fprintf stderr "LOCAL  PRAGMA: %s:%s=%s\n" f n v; *)
      add_pragma_to_definition !current_definition (f,n,v)
   in
   let end_of_rule () =
      let rule = List.rev !current_rule in
      add_rule !current_definition rule;
      current_rule := []
   in
   let end_of_rule_list () =
      (* nothing special to do here ?
         a following new_equation will do the job 
      *)
      ()
   in

   let parse_error (expcts: string) (tk: token) : unit =
      let (l,c) = Ylexer.curpos () in
      Printf.fprintf stderr "Error while parsing yacc, near line %d, char %d: %s expected, but get %s\n"
         l c expcts (string_of_token tk);
      exit 1
   in

   (* grammar parser *)
   let rec parse_grammar = function
   |  TK_end ->
      ()
   |  TK_pragma p -> 
      (* global pragma: it will be
         distributed to the following
         equations
      *)
      add_global_pragmas p;
      parse_grammar (lexer ())
   |  TK_ident i -> (
      skip TK_colon;
      (* init current left *)
      new_equation i;   
      match (lexer ()) with
      | TK_pragma p ->
         add_local_pragma p;
         parse_rulelist (lexer ());
         parse_grammar (lexer ()) 
      | tk ->
         parse_rulelist tk;
         parse_grammar (lexer ()) 
   )
   | tk -> parse_error "pragma or ident" tk
   (* rules parser *)
   and parse_rulelist tk =
      match parse_rule tk with
      | TK_bar -> 
         end_of_rule ();
         parse_rulelist (lexer ())
      | TK_semicol -> (
         end_of_rule ();
         end_of_rule_list ()
      )
      | tk -> parse_error "bar or semicolon" tk
   (* returns the reason of the ending *)
   and parse_rule (tk: token) : token =
      match tk with
      | TK_ident i ->
         add_to_rule i;
         parse_rule (lexer ()) 
      | TK_literal l ->
         add_to_rule l;
         add_literal l;
         parse_rule (lexer ())
      | TK_directive d when d = "%prec" ->
         (* %prec is the only supported directive
            it supposes to skip the following token
         *)
         let _ = lexer () in
         parse_rule (lexer ())
      | _ -> tk
   (* read and skip an expected token *)
   and skip expect = match lexer () with
      | tk when (tk = expect) -> ()
      | tk -> parse_error (string_of_token expect) tk
   in
   let _ = parse_grammar (lexer ())
   in
   (* finalyze:
      reverse def,
      reverse rule list in tab
   *)
   let reverse_rules _id def =
      def.def_rules <- List.rev def.def_rules;
   in
   Hashtbl.iter reverse_rules zegram.gr_tab;
   zegram.gr_def_list <- List.rev zegram.gr_def_list;
   zegram
)

let dump g =
   let do_symbol i =
      let rls = symbol_rules g i in
      let pgs = symbol_pragmas g i in
      let pprag (f,n,v) =
         Printf.printf "%s <-- %s:%s=%s\n" i f n v
      in
      List.iter pprag pgs;
		let pfx0 = Printf.sprintf "%s ::= " i in
		let pfx1 = Printf.sprintf "%s  |  " (String.make (String.length i) ' ') in
		let prule pfx rl = (
			Printf.printf "%s" pfx ;
			List.iter (function r -> Printf.printf " %s" r) rl ;
			Printf.printf "\n";
			pfx1 
		) in
		let _ = List.fold_left prule pfx0 rls in
		Printf.printf "\n"
   in
   List.iter do_symbol (def_symbols g)



