(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)
(*
V2.0 re-writing:
try to keep it as close as the old v1.1
*)

open Utils

(******* RULES TO (E)BNF *************)

(* Is it a list ?
In yacc, list should be terminal LEFT-recursive rules: in LALR
parsing, left-recursive means tail-recursive, that is, stack-free.
Indeed, one can write write right recursive rules, but that may cause
stack overflow.

A def "X = rl" is a list if each r is:
"X l" or "l" and X does not appear on l

*)
exception Not_relevant

(* util: longest common prefix *)
let lpfx l = (
	let res = match l with
	|	[] -> []
	|	h::t -> (
		let rec treatn p n = (
			match (p,n) with
			|	([], _)
			|	(_, []) -> []
			|	(x::p',y::n') -> (
				if (x = y) then x::(treatn p' n')
				else []
			)
		) in
		List.fold_left treatn h t
	) in
(*
Printf.fprintf stderr "LPFX OF %s = %s\n"
(string_of_string_list_list l)
(string_of_string_list res);
*)
	res
)
(* util: remove common prefix *)
let rmpfx p l = (
(*
Printf.fprintf stderr "RMPFX %s in %s\n"
(string_of_string_list p)
(string_of_string_list_list l);
*)
	let rec rmp p r = (
		match (p,r) with
		|	([],_) -> r
		|	(x::p',y::r') -> (
(*
Printf.fprintf stderr "  p=%s::%s  r=%s::%s\n"
x (string_of_string_list p')
y (string_of_string_list r');
*)
			assert (x = y);
			rmp p' r'
		)
		|	_ -> assert false
	) in
	List.map (rmp p) l
)

(*** rule_to_ebnf checks FIRST for left-rec definition *)
let rec rule_to_ebnf x rl = (
	(* try rec listr first *)
(* Printf.printf "CALL: rule_to_ebnf %s %s\n" x (string_of_string_list_list rl); *)
	let newdef =
		try check_for_star x rl
      (* why not check right rec ? *)
		with Not_relevant ->
		right_list_to_ebnf rl
	in
	(x, newdef)
) and
check_for_rec_def (rightrec:bool) (x:string) (rl: string list list) = try (
	(* accumulates a triple (base, ratio), such that
      x = x.(ratio) + base
		or raise Not_relevant
	*)
Verbose.exe (fun () ->
   Printf.fprintf stderr "\n*** check_for_rec_def\n";
   Printf.fprintf stderr "   %s\n"
      (if rightrec then "RIGHT REC" else "LEFT REC");
   Printf.fprintf stderr "   SYMB: %s\n" x;
   Printf.fprintf stderr "   DEF:  %s\n"
      (string_of_string_list_list rl);
);
	let treat_rule (base,ratio) r = (
		match r with
		|	[] -> ([]::base, ratio)
		|	y::r' -> (
			if (x = y) then (
				if List.mem x r' then raise Not_relevant
				else (base, r'::ratio)  
			) else (
				if List.mem x r then raise Not_relevant
				else (r::base,ratio)
			)
		)	
	) in
   (* if rightrec, reverse all rules in rl *)
   let rl = if rightrec then List.map List.rev rl else rl in
   let (base,ratio) = List.fold_left treat_rule ([],[]) rl in
   (* if it not failed but no ratio => not a list *)
   let _ = if ratio = [] then raise Not_relevant in 
   (* if rightrec reverse all ets within ratio and base *)
   (* in all case reverse ratio and base to "rebuild" user order *) 
   let (base,ratio) = if rightrec
      then (List.map List.rev base, List.map List.rev ratio)
      else (base, ratio)
   in
   let (base,ratio) = (List.rev base, List.rev ratio) in 

Verbose.exe (fun () ->
if rightrec then (
   Printf.fprintf stderr "    -> RIGHT REC DEF: %s = %s.%s | %s\n\n"
   x
   (string_of_string_list_list ratio) x
   (string_of_string_list_list base) 
) else (
   Printf.fprintf stderr "    -> LEFT REC DEF: %s = %s | %s.%s\n\n"
   x (string_of_string_list_list base)
   x (string_of_string_list_list ratio)
));
   (* optim: ratio U eps = base -> base = empty ?  *) 
   (base, ratio)
) with Not_relevant -> (
   Verbose.printf "   -> Not_relevant\n\n";
   raise Not_relevant
)
and
(* checks, recursively, for LEFT or RIGHT rescursive def *)
check_for_star (x:string) (rl: string list list) = (
   try (
      (* left ? *)
      let (prefix, ratio) = check_for_rec_def false x rl in
      (* x = prefix {ratio} *)
		Ebnf.make_seq [
			right_list_to_ebnf prefix;
			Ebnf.Star (right_list_to_ebnf ratio)
		]
   ) with Not_relevant -> (
      (* right ? *)
      let (suffix, ratio) = check_for_rec_def true x rl in
      (* x = {ratio} suffix *)
		Ebnf.make_seq [
(* here bug found by Michel Levy -> do not reverse ! *)
			Ebnf.Star (right_list_to_ebnf ratio);
			right_list_to_ebnf suffix
		]
   )

) and
(*** right_list_to_ebnf checks for common factors *)
factorize rl = (
	match rl with
	|	[]  -> ([],[],[])
	|	[r] -> ([],[r],[])
	|	_ -> (
		(* more than 2 choices, let's try to factorize *)
		let pfx = lpfx rl in
		let remain = rmpfx pfx rl in 
		let rev_remain = List.map List.rev remain in
		let rev_sfx = (lpfx rev_remain) in
		let center = List.map List.rev (rmpfx rev_sfx rev_remain) in
		let sfx = List.rev rev_sfx in
Verbose.exe (fun () ->
Printf.fprintf stderr "FACTORS OF %s :\n" (string_of_string_list_list rl);
Printf.fprintf stderr "%s -- %s -- %s :\n"
	(string_of_string_list pfx)
	(string_of_string_list_list center)
	(string_of_string_list sfx)
	;
);
		(pfx, center, sfx)
	)
) and
right_list_to_ebnf rlin = (
	(* contains eps ? *)
	let (rl, eps) = if (List.mem [] rlin) then (
		List.filter (function r -> (r <> [])) rlin,
		true 
	) else (
		rlin,
		false
	) in 
	(* factorize if possible ... *)
	let (pfx, center, sfx) = factorize rl in
	let facrl = if ( pfx <> []) || (sfx <> [] ) then (
		(* recursive call on center *)
		let bnf_center = right_list_to_ebnf center in
		(* let bnf_center = Ebnf.make_choice (List.map right_to_ebnf center) in *)
		Ebnf.make_seq [
			right_to_ebnf pfx;
			bnf_center;
			right_to_ebnf sfx;
		]	
	) else Ebnf.make_choice (List.map right_to_ebnf rl)
	in
	(* put in opt ? *)
	try (
		if(eps) then Ebnf.make_opt facrl
		else facrl
	) with _ -> (
		Printf.fprintf stderr "Error while computing ebnf of:\n";	
		Printf.fprintf stderr "  %s\n" (string_of_string_list_list rlin);
		Printf.fprintf stderr "--> unexpected Opt on:\n";	
		Printf.fprintf stderr "  %s\n" (Ebnf.string_of_ebnf (Ebnf.create ()) facrl);
		exit 1
	)
) and
right_to_ebnf r = (
	Ebnf.make_seq (List.map (function s -> Ebnf.Symb s) r)
)

(*
Input:
- a grammar Grammar.t,
- a (literal) token def list (string * string) list
- a (symbol) token def list (string * string) list
Returns :
- a Ebnf.t
*)
let f (g:Grammar.t)
      (tks: (string * string) list)
      (symbs: (string * string) list)
: Ebnf.t =

   (* init the tokens in Ebnf:
      Are treated as literal:
      - all the literal in Grammar.gr_lit_symbols
      - plus the extra tokens of tks
      Are treated as symbols: 
      - the extra tokens of symbs 
   *)
   List.iter (fun x -> Ebnf.add_terminal (x,x)) (Grammar.lit_symbols g);
   List.iter (fun x -> Ebnf.add_terminal x) tks;
   List.iter (fun x -> Ebnf.add_nonterminal x) symbs;

   let ebnf_g = Hashtbl.create 50 in
   let id_mode x = try (
      match Grammar.get_symbol_pragma g x ("ebnf","print") with
      | "short"-> Ebnf.Short
      | "expand"-> Ebnf.Expand
      | "ignore"-> Ebnf.Ignore
      | m -> (
         Printf.fprintf stderr "Error: unexepected ebnf:print mode \"%s\"\n" m;
         exit 1
      )
      ) with Not_found -> Ebnf.Normal
   in
   let treat_id x = (
      let rl = Grammar.symbol_rules g x in
      let (_,d) = rule_to_ebnf x rl in
      Hashtbl.add ebnf_g x (d, id_mode x)
   ) in
   List.iter treat_id (Grammar.def_symbols g);
   ebnf_g

