(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)
(* Set utilities *)

(* ====== PRINTABLE FUNCTIONAL TYPES UTILS =====*) 
(* Signature: ordered and printable types *)
module type XType =
   sig
      type t
      val compare: t -> t -> int
      val to_string: t -> string
   end;;

(* MODULE: ORDERED and PRINTABLE INT *)
module XInt : XType =
  struct
    type t = int
    let compare = Stdlib.compare
    let to_string = Stdlib.string_of_int
  end;;

(* MODULE: ORDERED and PRINTABLE STRING *)
module XString : XType =
  struct
    type t = string
    let compare = Stdlib.compare
    let to_string = fun x -> x
  end;;

(* FUNCTOR: SETS OVER ORDERED and PRINTABLE TYPES *)
module XSet (Elt: XType) =
   struct
      include Set.Make (
         struct
            type t = Elt.t
            let compare = Elt.compare
         end
      ) 
      let to_string: t -> string = fun x ->
         "{"^(String.concat ", " (List.map Elt.to_string (elements x))) ^"}"

   end;;

(* MODULE: PRINTABLE INT SET *)
module IntSet = XSet (
   struct
      type t = int
      let compare = Stdlib.compare
      let to_string = Stdlib.string_of_int
   end
);;

module StringSet = XSet (
   struct
      type t = string 
      let compare = Stdlib.compare
      let to_string = (fun x -> x)
   end
);;

(* FUNCTOR: MAPS BETWEEN ORDERED and PRINTABLE TYPES *)
module XMap (Key: XType) (Val: XType) =
   struct
      module M = Map.Make (
         struct
            type t = Key.t
            let compare = Key.compare
         end
      ) 
      open M 
      type t = Val.t M.t
      let cardinal = M.cardinal
      let empty = M.empty
      let singleton = M.singleton
      let add = M.add
      let find = M.find
      let fold = M.fold
      let map = M.map
      let mem = M.mem
      let is_empty = M.is_empty
      let remove = M.remove
      let filter = M.filter
      let merge = M.merge
      let to_string ?(mapsto:string=" -> ") (x:t) : string =
         let bs (k, v) = "("^(Key.to_string k)^mapsto^(Val.to_string v)^")" in
         "{"^(String.concat ", " (List.map bs (bindings x))) ^"}"
      let of_list: (Key.t * Val.t) list -> t = fun lst ->
         List.fold_left (fun a (k,v) -> add k v a) empty lst 
(*
      let to_string: Val.t t -> string = fun x ->
         let bs (k, v) = "("^(Key.to_string k)^" -> "^(Val.to_string v)^")" in
         "{"^(String.concat ", " (List.map bs (bindings x))) ^"}"
      let of_list: (Key.t * Val.t) list -> Val.t t = fun lst ->
         List.fold_left (fun a (k,v) -> add k v a) empty lst 
*)
   end;;

(* MODULE: INT 2 STRING MAP *)
(* ??  module Int2String = XMap (struct include XInt end) (struct include XString end);; *)
module Int2String = XMap
(
   struct
      type t = int
      let compare = Stdlib.compare
      let to_string = Stdlib.string_of_int
   end
)
(
   struct
      type t = string
      let compare = Stdlib.compare
      let to_string = fun x -> x
   end
);;

module Int2Int = XMap
(
   struct
      type t = int
      let compare = Stdlib.compare
      let to_string = Stdlib.string_of_int
   end
)
(
   struct
      type t = int
      let compare = Stdlib.compare
      let to_string = Stdlib.string_of_int
   end
);;

module Int2IntSet = XMap
(
   struct
      type t = int
      let compare = Stdlib.compare
      let to_string = Stdlib.string_of_int
   end
)
(
   struct
      type t = IntSet.t
      let compare = IntSet.compare
      let to_string = IntSet.to_string
   end
);;


(* ====== MISC PRINTERS UTILS =====*) 

let string_of_list
   ?(obr: string = "[")
   ?(cbr: string = "]")
   ?(sep: string = "; ")
   (soa: 'a -> string) (al: 'a list) = 
   obr^(String.concat sep (List.map soa al))^cbr

let string_of_string_list
   ?(obr: string = "[")
   ?(cbr: string = "]")
   ?(sep: string = "; ")
   sl = string_of_list ~obr:obr ~cbr:cbr  ~sep:sep (fun x->x) sl

let string_of_int_list
   ?(obr: string = "[")
   ?(cbr: string = "]")
   ?(sep: string = "; ")
   sl = string_of_list ~obr:obr ~cbr:cbr ~sep:sep string_of_int sl

let string_of_string_list_list
   ?(obr1: string = "[")
   ?(cbr1: string = "]")
   ?(sep1: string = "; ")
   ?(obr2: string = "[")
   ?(cbr2: string = "]")
   ?(sep2: string = "; ")
   sll =
   string_of_list ~obr:obr1 ~cbr:cbr1 ~sep:sep1
      (string_of_string_list ~obr:obr2 ~cbr:cbr2 ~sep:sep2) sll


let rec string_of_assocs = function
   (k,v)::[] -> k^"="^v
|  (k,v)::t -> k^"="^v^" "^(string_of_assocs t)
|  [] -> ""

let string_of_option (soa: 'a -> string) (ao: 'a option) = 
   match ao with
   | None -> "None"
   | Some a -> "Some("^(soa a)^")" 

(* ====== MISC HASH UTILS =====*) 
let string_of_hashtab (sok: 'a -> string) (soe: 'b -> string) (abtbl: ('a,'b) Hashtbl.t) =
   let soab k v acc = ((sok k)^" -> "^(soe v)^";")^acc in
   "["^(Hashtbl.fold soab abtbl "")^"]"

let hash_content ht =
   Hashtbl.fold (fun k v a -> (k,v)::a) ht []

let sorted_hash_content ht =
   let c = hash_content ht in
   List.fast_sort Stdlib.compare c

(* foo to bar list hashtables *) 
(* get elt in 'key -> val list' Hashtbl
   returns [] if Not_found 
*)
let find_hash_list (tab: ('a, 'b list) Hashtbl.t) (k: 'a) : 'b list =
try Hashtbl.find tab k with Not_found -> []

(* add elt in 'key -> val list' Hashtbl *)
let add_hash_list (tab: ('a, 'b list) Hashtbl.t) (k: 'a) (v: 'b) =
   let oldv = try
      Hashtbl.find tab k
   with Not_found -> []
   in
   Hashtbl.replace tab k (v::oldv)

(* JUST A TEST
type toto = int * string
module TotoSet = Set.Make (
   struct
      type t = toto
      let compare = Stdlib.compare
   end
)
*)

let rec iter_sep (f: 'a -> unit) (g: unit -> unit) (l: 'a list) : unit =
   match l with
   | [] -> ()
   | [x] -> f x
   | x::l' -> f x; g (); iter_sep f g l'

let iteri_sep (f: int -> 'a -> unit) (g: unit -> unit) (l: 'a list) : unit =
   let rec _iteri i l =
      match l with
      | [] -> ()
      | [x] -> f i x
      | x::l' -> f i x; g (); _iteri (i+1) l'
   in
   _iteri 0 l

let rec make_list (i: int) (x:'a) : 'a list =
   if i <= 0 then [] else x::(make_list (i-1) x)


(* PHYSICAL-IDENTITY HASH TABLE
usage example:

module Stbl = PhyHtbl (
   struct
      type t = string 
   end
);;
*)

module type PhyKey =
sig
   type t
end;;

module PhyHtbl (K: PhyKey) =
struct
   let phy_hash (x:'a) = (
      if Obj.is_block (Obj.repr x) then
         Hashtbl.hash (Nativeint.of_int (Obj.magic x)) (* magic *)
      else
         invalid_arg "Can only find address of boxed values."
   )
   include Hashtbl.Make(
      struct
         type t = K.t
         let equal = (==)
         let hash = phy_hash
      end
   )
end;;
