(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)


(* open Utils *)

module V = Verbose

let _VERSION = Printf.sprintf "3.0-r%s" (SrcRevision.code)
let _TOOLNAME = "yacc2latex"
let usage_msg = Printf.sprintf "%s version %s\nusage: %s <option> | %s -help"
_TOOLNAME _VERSION
_TOOLNAME _TOOLNAME

(*

3.0:
   - add stdin/stdout support
   - add degub flags
   - change Pervasives to Stdlib 

2.2: bug in right-rec  fixed, add Verbose

2.0:
- pure-ocaml (almost) complete re-write, no more c code !
- in particular: intermediate "rule" format is no longer used
  replaces by Grammar + Ylexer modules

1.1:
- new feature: regognize righ-rec defs as kleene-star + suffix
  (however its a bad idea in yacc !)
1.0:
- fixed bug: non-terminal symbols can start with _, contain '.' etc
  actually, they match [_\.0-9a-zA-Z]+
  (maybe accents, but not sure ...
*)


(* Mini man ... *)
let miniman = Miniman.text

(* arguments *)
let print_version () = (
	Printf.printf "%s" _VERSION;
   exit 0
)

let dbg = V.get_flag "main"

let arg_list = ref []

(* clean up md *)
let sminiman = Str.global_replace (Str.regexp "#+ ") "" miniman
let sminiman = Str.global_replace (Str.regexp "```") "" sminiman

let rec print_man () = (
	prerr_string "-------------------------------------\n";
	prerr_string ("   "^ _TOOLNAME^" version "^_VERSION^" \n");
	prerr_string "             Short Manual            \n";
	prerr_string "       Jun 2012 (rev May 2014)       \n";
	prerr_string "          Pascal Raymond             \n";
	prerr_string " http://www-verimag.imag.fr/~raymond \n";
	prerr_string "-------------------------------------\n";
	Arg.usage !arg_list "";
	prerr_string sminiman;
	exit 0
)
and print_man_md () = (
   prerr_string ("# Short Manual\n");
   prerr_string ("- **Version**: "^_VERSION^" \n");
   prerr_string  "- **Pascal Raymond** [(Home Page)](http://www-verimag.imag.fr/~raymond)\n";
   prerr_string "\n# Usage\n";
   prerr_string "```\n";
	Arg.usage !arg_list usage_msg;
   prerr_string "```\n";
	prerr_string miniman;
	exit 0
)
and _INFILE = ref ""
and _STDIN = ref false
and _OUTFILE = ref ""
and _STDOUT = ref false
and _TOKFILE = ref ""
and _DUMPL = ref false
and _DUMPY = ref false
and _DUMPG = ref false
and _MAKEMAIN = ref false
and set_infile = function (x:string) -> (
   if ( !_INFILE = "" ) then (
      _INFILE := x
   ) else (
      Arg.usage
         !arg_list
         ("unexpected argument \""^x^"\"\n"^usage_msg) ;
      exit 1
   )
)
and bad_arg x =
      Arg.usage
         !arg_list
         ("unexpected argument \""^x^"\"\n"^usage_msg) ;
      exit 1
let _ =  arg_list :=
let dbgflags = V.flag_list () in
[
("-m",       Arg.Unit print_man,       "      :  print a short manual");
("-md",      Arg.Unit print_man_md,    "      :  same in markdow format");
("-t",       Arg.Set_string  _TOKFILE, "<file>:  load token def file");
("-o",       Arg.Set_string  _OUTFILE, "<file>:  set out file (default infile.tex)");
("-os",      Arg.Set _STDOUT,          "      :  write output to stdout");
("--",       Arg.Set _STDIN,           "      :  read input from stdin");
("-main",  Arg.Set _MAKEMAIN,             "   :  gen a sample main latex file)");
("-testl",   Arg.Set _DUMPL,               "  :  dump the yacc tokens and stop");
("-testy",   Arg.Set _DUMPY,               "  :  dump the yacc grammar and stop");
("-testg",   Arg.Set _DUMPG,               "  :  dump the ebnf grammar and stop");
("-version", Arg.Unit print_version,         ":  print version and stop");
("-v", Arg.Unit V.on,         ":  set verbose mode");
("-dbg", Arg.Symbol (dbgflags,
   (function s -> let x = V.get_flag s in V.set_flag x)), ": set debug flag");
]

let latex_out () =
   (* a "filename" is forced even when stdout is set 
      which serves as 'basename' in case -main  is selected
   *)
   let default_name = match !_INFILE with
      | "" -> "_out_.tex"
      | inf -> (Filename.chop_extension (Filename.basename inf))^".tex"
   in
   match (!_OUTFILE, !_STDOUT) with
   | ("", true) ->
      V.printf ~f:dbg "DBG: writing to stdout\n";
      (stdout, default_name)
   | (_, true) ->
      Arg.usage
         !arg_list
         ("ambiguous output\n"^usage_msg) ;
      exit 1
   (* stdout not set ... *)
   (* out file NOT given: 
      if _STDIN, take stdout, else open outf_dflt
   *)
   | ("", _) when !_STDIN ->
      V.printf ~f:dbg "DBG: writing to stdout\n";
      (stdout, default_name)
   | ("", _) ->
      V.printf ~f:dbg "DBG: writing to \"%s\"\n" default_name;
      (open_out default_name, default_name)
   (* out file given: take it *)
   | (f, _) ->
      V.printf ~f:dbg "DBG: writing to \"%s\"\n" f;
      (open_out f, f)

let latex_main_out outf =
   (* include basename of out*)
   let toincl = Filename.basename outf in
   let moutf = (Filename.chop_extension outf)^"_main.tex" in
   (open_out moutf, toincl) 
   

(* search for "group" specificatons *)
let mk_groups g =
   let grtab = Hashtbl.create 20 in 
   let grlist = ref [] in
   let add2group gr id =
      let idl = try
         Hashtbl.find grtab gr
      with Not_found -> (
         grlist := gr::!grlist;
         []
      ) in
      Hashtbl.replace grtab gr (id::idl) 
   in
   let orphans = ref [] in
   let treat_id id =
      try
         let gr = Grammar.get_symbol_pragma g id ("ebnf","group") in
         add2group gr id
      with Not_found -> 
         orphans := id::!orphans
   in
   List.iter treat_id (Grammar.def_symbols g);
   let mkg g =
      let idl = Hashtbl.find grtab g in
      (g, List.rev idl)
   in
   match (!orphans, !grlist) with
   | (_, []) -> []
   | ([], gl) -> List.map mkg (List.rev gl)
   | (_ol, gl) -> List.map mkg (List.rev ("orphan-rules"::gl))

(* read a tok def file ...
   CONVENTION: a "token" starting with %
   is treated (and will be displayed) as a symbol
   otherwise, they are treated and displayed as literal strings
   
*)
type token_spec =
| Lit of (string * string)
| Ide of (string * string)

let read_pair l = (
   let mkc x y = (x,y) in
   (* de preference on scane %s %S *)
   let (tk,v) = try
      Scanf.sscanf l "%s %S" mkc
   with _ -> (
      Scanf.sscanf l "%s %s" mkc
   )
   in
V.printf "token \"%s\" = \"%s\"\n%!" tk v;
   match tk with
   | "" -> None
   | _ -> (
      match tk.[0] with
      | '%' -> Some (Ide (String.sub tk 1 ((String.length tk) - 1), v))
      | _ -> Some (Lit (tk,v)) 
   )
)

let read_tok_file tkfile = try (
   let tch = open_in tkfile in
   let tks = ref [] in
   let symbs = ref [] in
   let rec treat_line () = try (
      let l = input_line tch in
      (* Printf.printf "line=%s\n" l; *)
      let _ = match read_pair l with
      | None -> ()
      | Some Lit x -> tks := x::!tks
      | Some Ide x -> symbs := x::!symbs
      in
      treat_line ()
   ) with End_of_file -> [] 
   in
   let _ = treat_line () in
   (!tks, !symbs)

) with  Sys_error m -> (
      Printf.fprintf stderr "ERROR, %s\n" m;
      exit 1
         (* "Can't open file \"%s\"\n" tkfile *)
)

let main () = (
	Arg.parse !arg_list set_infile usage_msg;

(* Printf.printf "_DUMPY=%b\n" !_DUMPY; *)
(* Printf.printf "_DUMPG=%b\n" !_DUMPG; *)
   let inc = match (!_INFILE,!_STDIN) with
   | ("",false) ->
	   Arg.usage !arg_list usage_msg;
      exit 0
   | ("",true) ->
      V.printf ~f:dbg "DGB: reading from stdin\n";
      stdin
   | (_f,true) -> 
      Arg.usage
         !arg_list
         ("ambiguous input\n"^usage_msg) ;
      exit 1
   | (f,false) ->
      V.printf ~f:dbg "DBG: reading from \"%s\"\n" f;
      open_in f
   in
   let get_token = Ylexer.create inc in
   if !_DUMPL then (
      let rec iter_tokens () = try (
         let tk = get_token () in
         Printf.printf "%s\n" (Ylexer.string_of_token tk);
         iter_tokens ()
      ) with End_of_file -> ()
      in
      iter_tokens ()
   ) else ( 
      let g = Grammar.of_yacc get_token in
	   if !_DUMPY then (
         Grammar.dump g
      ) else (
         let (tks,symbs) = if (!_TOKFILE <> "")
            then read_tok_file !_TOKFILE   
            else ([],[])
         in 
         let ebnf_g = Grammar2Ebnf.f g tks symbs in 
         let ll = Grammar.def_symbols g in
         let groups = mk_groups g in
		   if !_DUMPG then (
			   Ebnf.dump ebnf_g ll groups;
			   exit 0
		   ) else (
            let (outc,outf) = latex_out () in 
			   Ebnf.to_latex outc ebnf_g ll groups;
            if (!_MAKEMAIN && (outc != stdout)) then (
               let (moutc, toincl) = latex_main_out outf in
               Ebnf.make_latex_top moutc toincl
            )
		   )
      )
	)
)

let _ = main ()



