(***********************************************************************
yacc2latex

Pascal Raymond, Verimag

http://www-verimag.imag.fr/~raymond/

Copyright 2012,2015 Pascal Raymond,
Centre National de la Recherche Scientifique

    This file is part of yacc2latex.

    yacc2latex is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    yacc2latex is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with yacc2latex.  If not, see <http://www.gnu.org/licenses/>.

****************************************************************************)

(*

Abstract input file organization, 
valid for classical C/C++ yacc/bison files
and for ocaml mly files

-------------
%{
host language code
%}
Grammar declarations
%%
Grammar rules
%%
Additional host language code
-------------

Note: Grammar decl/rules may contain comment (C-like i.e. /* ... */)

What we do :
- ignore anything until the first "^%%" line
- Parse the rule part until the following ^%% (or EOF)

valid "tokens" in the rule part:
- yacc ident  (slightly mode general than c ident)
- literal character token (e.g. ';')
- literal string token (e.g. "<=" )
- separators ":" "|" and ";"
- host code "{ ... code ... }"
- comment (or pragma) /* ... text ... */

*)

open Printf

(*
according to bison manual:
"Symbol names can contain letters, digits (not at the beginning), underscores and periods.
Periods make sense only in nonterminals."
Note: we don't really care about identifiers starting with a digit... 
-> extended alpha num strings ...
*)
let isxalnum (c:char) : bool =
match c with
| 'A' .. 'Z'
| 'a' .. 'z'
| '0' .. '9'
| '_'
| '.' -> true
| _ -> false

(* ansi c ref The standard white-space characters are the following:
space, tab, carriage-return, newline, vertical tab, and form-feed
9  011   09 00001001 HT &#009;      Horizontal Tab
10 012   0A 00001010 LF &#010;      Line Feed
11 013   0B 00001011 VT &#011;      Vertical Tab
12 014   0C 00001100 FF &#012;      Form Feed
13 015   0D 00001101 CR &#013;      Carriage Return
*)
let isspace = function
| ' '
| '\x09' .. '\x0D' -> true
| _ -> false

(* isprint : only basic 7bit printable codes, range 0x20 .. 0x7E *)
let isprint = function
|  '\x20' .. '\x7E' -> true
| _ -> false

let lit_of_char c =
   if isprint c then sprintf "'%c'" c
   else sprintf "'\\x%2X'" (int_of_char c)

(* TOP FUNCTION *)
type token =
(* separator *)
| TK_colon
| TK_bar
| TK_semicol
(* immediate char *)
| TK_literal of string
(* ident *)
| TK_ident of string 
(* yacc directive *)
| TK_directive of string 
(* pragmas *)
| TK_pragma of (string * string * string)
(* end marker -> ease the parsing
   any subsequent call of get_token will
   raise End_of_file exception !
*)
|  TK_end

let string_of_token = function
| TK_colon -> ":"
| TK_bar   -> "|"
| TK_semicol -> ";"
| TK_literal s -> "L="^s
| TK_ident   s -> "I="^s
| TK_directive    s -> "Y="^s
| TK_pragma  (x,y,z) -> sprintf "P=(%s,%s,%s)" x y z
| TK_end  -> "%%"

(* -------------------------------------------
tokenizer:
"object-like" (but without ocaml objects !)
input: an input stream
output: a function that iterates the tokens, and
   raises End_of_file when finished
 ------------------------------------------- *)
type t = unit -> token

(* ugly -> externalize charcpt and linecpt for error localization ...
   (Lxm stuff is much better ... )
*)

let charcpt = ref 0
let linecpt = ref 1

let curpos () = (!linecpt, !charcpt)

let create (inc: in_channel) : t = (

   (* INPUT FILE ACCESS char by char
      cheat: use null-char as eof marker,
      (more confortable than dealing with exceptions)
      + look_forward : 1 char forward ONLY
   *)
   let (eof:char) = '\x00' in
   let cbuff : char option ref = ref None in
   (* let charcpt = ref 0 in *)
   (* let linecpt = ref 1 in *)
   charcpt := 0;
   linecpt := 1;
   let getc () : char = (
      match !cbuff with
      |  Some c ->
         cbuff := None;
         c
      |  None -> try (
            let c =  input_char inc in
            if c = '\n' then (
               incr linecpt;
               charcpt := 0
            );
            incr charcpt;
            c
         ) with End_of_file -> eof
   ) in
   let ungetc (c: char) : unit = (
      match !cbuff with
      |  Some _ ->
         fprintf stderr "unget facility limited to a 1 char buffer\n";
         assert false
      |  None ->
         cbuff := Some c
   ) in
   let skipc () : unit =
      let _ = getc () in ()
   in
   (* normally seekc MUST be equivalent to "let c = getc() in ungetc c; c" *)
   let seekc () : char = (
      match !cbuff with
      |  Some _ ->
         fprintf stderr "seek facility limited to a 1 char buffer\n";
         assert false
      |  None -> 
         let c = getc () in
         cbuff := Some c;
         c 
   ) in
   (* skip spaces *)
   let rec skip_spaces () =
      let c = getc () in
      if isspace c then skip_spaces ()
      else c
   in

   (* INPUT FILE ACCESS line by line 
      global error if End_of_file
   *)
   let get_line () = try
      let l = input_line inc in
      incr linecpt;
      l
   with End_of_file ->
      fprintf stderr  "Unexpected end of file nead line %d\n" !linecpt;
      exit 1
   in
   (* (re) use the same buffer when needed *)
   let buff = Buffer.create 512 in

   (* PARSING 1: ignore host code *)
   (* roughly read line by line until the first %% *)
   let gramrg = Str.regexp "[ \t]*%%[ \t]*" in
   let rec goto_grammar () = 
      let l = get_line () in
      if (Str.string_match gramrg l 0) then ()
      else goto_grammar ()
   in

   (* PARSING 2: get rule token *)
   let unexpected w c =
      fprintf stderr "Error: while %s, unexpected char %s (at %d:%d)\n"
         w (lit_of_char c) !linecpt !charcpt ;
      exit 1
   in
   let finished = ref false in
   let rec get_token () = (
      if !finished then raise End_of_file;
      (* first non space char decides ... *)
      let c = skip_spaces () in
(* Printf.printf "****c=%c (%d)\n" c (Char.code c); *)
      match c with
      | ':' -> TK_colon
      | '|' -> TK_bar
      | ';' -> TK_semicol
      | '\'' -> get_literal_char ()
      | '"' -> get_literal_string ()
      | '{' -> treat_code ()
      | '/' when (seekc () = '*') -> (
         skipc (); treat_comment ()
      )
      | '%' when  (seekc () = '%') -> (finished:=true; TK_end)
      | '%' -> get_directive ()
      | z when  (z = eof) -> (finished:=true; TK_end)
      | _ -> (
         if isxalnum c then get_identifier c 
         else unexpected "searching token" c
      )
      
   )
   (* treat_comment: the initial "/*" is consumed,
      - check for pragma, if yes keep it
      - consume till the next */
      - if a pragma was kept, return it
      - else rec. call get_token

      WARNING DOES NOT SUPPORT NESTED COMMENTS
   *)
   and treat_comment () = (
      (* fill buffer until end, new line, or 512 char max *)
      Buffer.reset buff;
      let rec consume dobuff = (
         let c = getc () in
         if (c = '*') && (seekc () = '/') then skipc () 
         else if c = eof then
            unexpected "parsing comment" c
         else (
            let dobuff' = dobuff &&
               (c != '\n') &&
               (Buffer.length buff < 512)
            in
            if dobuff' then Buffer.add_char buff c;
            consume dobuff'
         )
      ) in
      consume true;
      let buff = Buffer.contents buff in
      let prg = Str.regexp "^.*ebnf:\\([^=]*\\)=\\([^ ]*\\)" in
      if Str.string_match prg buff 0 then (
         let name = Str.matched_group 1 buff  in
         let value = Str.matched_group 2 buff in
         TK_pragma ("ebnf", name, value)
      ) else
      get_token () 
   ) 
   (* treat_code: initial brace is sconsumed
      (very) partial support of host language:
      check for well-balenced pairs {}
   *)
   and treat_code () = 
      let cpt = ref 1 in
      let rec consume_braces () =
         if !cpt = 0 then ()
         else
            let _ = match getc () with
            | '}' -> decr cpt
            | '{' -> incr cpt
            | z when (z = eof) -> 
               unexpected "ignoring code" z
            | _ -> ()
            in consume_braces ()
      in
      consume_braces ();
      get_token ()
   (* get_identifier: the already consumed first char is given 
   *)
   and get_identifier c =
      Buffer.reset buff;
      Buffer.add_char buff c;
      let rec fill () =
         let c = getc () in
         if isxalnum c then (
            Buffer.add_char buff c;
            fill ()
         ) else ungetc c
      in
      fill ();
      TK_ident (Buffer.contents buff)
   (* get_directive: the initial % is consumed
   *)
   and get_directive _c =
      Buffer.reset buff;
      Buffer.add_char buff '%';
      let rec fill () =
         let c = getc () in
         if isxalnum c then (
            Buffer.add_char buff c;
            fill ()
         ) else ungetc c
      in
      fill ();
      TK_directive (Buffer.contents buff)
   (* get_literal_char,
      get_literal_string:
      the starting symbol has been consumed 
      check for escaped \ chars 
   *)
   and get_literal_char () =
      Buffer.reset buff;
      let rec fill () =
         let c = getc () in
         let sc = seekc () in
         match c with
         | '\'' -> ()
         | '\\' when (sc = '\'') ->
            Buffer.add_char buff (getc ()); fill ()
         | '\\' when (sc = '\\') ->
            Buffer.add_char buff (getc ()); fill ()
         |  z  when (z = eof) ->
            unexpected "parsing literal" z
         | _ -> Buffer.add_char buff c; fill ()
      in
      fill ();
      TK_literal (Buffer.contents buff)
   and get_literal_string () =
      Buffer.reset buff;
      let rec fill () =
         let c = getc () in
         match c with
         | '"' -> ()
         | '\\' when (seekc () = '"') ->
            Buffer.add_char buff (getc ()); fill ()
         |  z  when (z = eof) ->
            unexpected "parsing literal" z
         | _ -> Buffer.add_char buff c; fill ()
      in
      fill ();
      TK_literal (Buffer.contents buff)
      
   in
   (* Prepare inc for iterating on tokens *)
   goto_grammar ();
   (* Returns get_token *)
   get_token
)
